! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!+
! Module: proton_mod
!
! Authors:  Brendan Perry - NASA/GSFC (brendan.h.perry@nasa.gov)
!           (BP) comprehensive rewrite and modularization for SAS-19
!           Steven L. Snowden - NASA/GSFC - original algorithm
!           
! 
! Description: Extended Source Analysis Software
!
! Comments:
!
! $Id: proton_mod.f90,v 1.3 2009/02/26 16:39:45 aibarra Exp $
!-
!  Version 5.0  09 November 2012
!  Version 6.0/0.1  01 March 2020 (rewrite by BP, added SAS DAL/CAL calls)
!  Version 0.11 20 April 2020 (reconciliation b/w old esas version numbers)
!
!  Casts a map of the soft proton background using the fitted
!  spectral parameters for a power law (works) and broken power
!  law (in progress).  Soft proton detector maps developed by
!  K. Kuntz are used for the distributions. This new version
!  supercedes the proton within the old esas package. 
!
!  Usage:
!
!   proton objimage=1S001 ccd1=1 ccd2=1 ccd3=1 ccd4=1 ccd5=1 ccd6=1 ccd7=1 elow=750 ehigh=1250 spectrumcontrol=1 pindex=0.441324 pnorm=1.7351400E-02
!
!     imagefile - Object Image (dataset)
!     specfile  - Object Spectral file (dataset)
!     withsky   - Perform det to sky transformation with esky2det/rotdet2sky?
!     spmapdet  - Output soft proton bkg map in det coords, defaults
!     spmapsky  - Output soft proton bkg map in sky coords, defaults
!     spectrol  - spectrum control (1=powerlaw,2=broken power law)
!     elow     - lower energy band limit in eV
!     ehigh    - upper energy band limit in eV
!     ccds     - CCDs/Quads to select
!     pindex   - power law index              (speccontrol=1) from Xspec
!     pnorm - power law normalization         (speccontrol=1) from Xspec
!     bindl - soft power law index            (speccontrol=2) from Xspec
!     break - power law break energy (in keV) (speccontrol=2) from Xspec
!     bindh - hard power law index            (speccontrol=2) from Xspec
!     bnorm - broken power law normalization  (speccontrol=2) from Xspec
!
module proton_mod

  use dal
  use cal
  use param
  use gofutils_mod
  use dataFilter
  use errorHandling
  use sasdal
  use gofio_mod

  implicit none

  private

  character(len=*), parameter :: &
       objimgpar    = "imagefile", &     ! was: objfile from 'prefix'
       objspecpar   = "specfile", &      ! was: spec_name
       withskypar   = "withsky", &       ! create sky coord output SP map?
       outdetpar    = "spmapdet", &      ! output SP bkg map in det coords
       outskypar    = "spmapsky", &      ! output SP bkg map in sky coords
       spectrolpar  = "speccontrol", &   ! was: spectrumcontrol 1 or 2
       elowpar      = "elow", &          ! low energy bound eV 
       ehighpar     = "ehigh", &         ! high energy bound eV
       ccdspar      = "ccds", &          ! replaced ccd1, ccd2, etc
       pindexpar    = "pindex",&         ! power law index
       pnormpar     = "pnorm",&          ! power law normalization
       bindlpar     = "bindl",&          ! broken power law low index
       bindhpar     = "bindh",&          ! broken power law high index
       bbreakpar    = "bbreak",&         ! broken power law break energy
       bnormpar     = "bnorm"            ! broken power law normalization

  public proton

contains

  !---------------------------------------------------------------------
  subroutine proton
    !---------------------------------------------------------------------

    type(DataSetT) :: inIMG_S, inSPEC_S, outSPdetmap_S, outSPskymap_S, outOBJIMG_S
    type(ArrayT) :: inIMG_A, outSPmap_A, out_A
    type(TableT) :: inSPEC_T, inIMG_T
    !    type(ColumnT) :: channelS
    character (len=2) :: inst, instO, instS, cc, bandd, &
         bands(6) = (/ 'B1','B2','B3','B4','B5','B6' /)
    character (len=5) :: celow, cehigh, expid, expidO, expidS
    character (len=6) :: filtO, filtS, &
         filter(3) = (/ 'THIN  ','MEDIUM','THICK ' /) 
    character (len=10) :: prefix, obsid, obsidO, obsidS, ctype(2), cunit(2), radecsys
    character (len=12) :: str_crval1, str_crval2, str_detx, str_dety, estring
    character (len=18) :: type, suffix, logfile
    character (len=25) :: smode, smodeO, smodeS, submode
    character (len=80) :: line
    character (len=256) :: inIMG_F, inSPEC_F, outSPskymap_F, &
         outSPdetmap_F, outOBJIMG_F
    character (len=1024) :: command
    integer, dimension(2) :: outIMGdim
    integer :: chansobs, instnum, typeF, nRowsO, nRowsS, revolut, revolutO, &
         revolutS, spectrol, inChans, spXsize, spYsize, ff, &
         elowO, elowS, ehighO, ehighS, frmtime, crap, reason
    integer :: elow, ehigh, iext, i, i1, i2, ib, ibb(7), ih, ios, ii, ii1, &
         ii2, iii, il, j, jj, k, kk, lun, nbands, n, nCCDs, typeA
    integer(kind=int16), dimension(:,:), pointer :: aDI16, inIMG_IP
    integer(kind=int32), dimension(:,:), pointer :: outSPmap_IP, aDI32, &
         inIMG_JP
    ! check channels for mos AND pn ???!!!
    integer, parameter :: specchans = 2400, flareX = 780, flareY = 780 ! cal file
    integer, parameter :: xSize = 900, ySize = 900 ! hardcode forever???!!!
    integer :: status = 0
    logical :: doband(6), ccds(7) ! bhp: one array for all ccds/quads needed
    logical :: powerlaw, withsky, i_ok, warndie, withlong

    real(single), dimension(:,:), pointer :: outSPmap_RP, flaremap_RP, aDR32
    real(double), dimension(:,:), pointer :: aDR64
    real(single), dimension(:,:), pointer :: temp_real32, inIMG_RP
    real :: tempp
    real :: bindl, break, bindh, bnorm, bnormh, bnorml, &
         eh, el, exposen, pindex, pnorm, SPrate, rind, rn, &
         scale, sphb, splb, temp
    real(kind=double) :: crval(2), cdelt(2), crpix(2), crota(2), &
         dxcen, dycen, ra_pnt, dec_pnt, pa_pnt
    real :: eLo(6) = (/300.0, 750.0, 1250.0, 2000.0, 4000.0, 8000.0/) ! old elim
    real :: eHi(6) = (/749.0, 1249.0, 1999.0, 3999.0, 7999.0, 12000.0/)
    real :: backO, areaO, expoO, backS, areaS, expoS, scaleFactor(6)
    !
    backO = 0.
    areaO = 0.
    expoO = 0.
    backS = 0.
    areaS = 0.
    expoS = 0.
    powerlaw = .FALSE.
    withlong = .FALSE.
    ccds = .FALSE.
    !
    ! If the img and spec files are no good, may as well quit early! (BP)
    ! should spec file be eliminated (get EXPOSURE from img file)??? (BP)
    !
    inIMG_F = stringParameter(objimgpar) ! was from prefix
    write(errstr,*) "FOV Image: ",trim(adjustl(inIMG_F))
    call message(AppMsg,Always) ! change to Noisy later!!!
    ! open and read input image file inIMG_F
    status = 0
    warndie = .TRUE. ! croak if not found or unreadable
    ! these below are in goflib/gofio_mod
    call check_open_set(inIMG_F,inIMG_S,1,status,warndie) 
    call read_2Darray(inIMG_S,inIMG_A,inIMG_RP,aDR64,aDI16,inIMG_JP,&
         spXsize,spYsize,typeA)
    call CAL_setState(inIMG_F)
    outIMGdim(1) = spXsize
    outIMGdim(2) = spYsize
    ! template image must be 900x900 (for mos AND pn???!!!)
    !    if ((spXsize .ne. xSize) .or. (spYsize .ne. ySize)) then
    !       write(errstr,*)"IMAGE pixels must be ",xSize," by ",ySize      
    !       call error("badIMGpixSize",trim(errstr))
    !    endif
    inIMG_JP => int32Array2Data(inIMG_A) ! FOV image array
    !
    ! extract basic keywords from input image file inobjIMG_F
    warndie = .TRUE.     ! see read_header_info in gofio_mod for behavior
    typeF = 3 ! 1 = evli, 2 = spec, 3 = image, instO = M1/M2/PN below
    call read_header_info(typeF,inIMG_S,inIMG_T,out_A,instO,obsidO,&
         expidO,smodeO,frmtime,filtO,revolutO,backO,areaO,expoO,elowO,ehighO,&
         nRowsO,warndie)
    !
    ! open and read input spectral file inobjSPEC_F
    inSPEC_F = stringParameter(objspecpar) ! was spec_name
    write(errstr,*) "FOV Spectrum: ",trim(adjustl(inSPEC_F))
    call message(AppMsg,Always) ! change to Noisy later!!!
    !
    status = 0
    warndie = .TRUE. ! croak if not found or unreadable
    call check_open_set(inSPEC_F,inSPEC_S,1,status,warndie)
    ! extract basic keywords from input spectral file inobjSPEC_F
    inChans = specchans ! hardcoded what *should* it be is 2400 always good? 
    warndie = .TRUE. ! see read_pha_header in gofio_mod for behavior
    typeF = 2 ! 1 = evli, 2 = spec, 3 = image
    call read_header_info(typeF,inSPEC_S,inSPEC_T,out_A,instS,obsidS,&
         expidS,smodeS,frmtime,filtS,revolutS,backS,areaS,expoS,elowS,&
         ehighS,nRowsS,warndie)
    if ((trim(instS) .ne. trim(instO)) .or. (trim(obsidS) .ne. trim(obsidO)) &
         .or. (trim(expidS) .ne. trim(expidO)) .or. (revolutS .ne. revolutO)) then
       call error("IMGneSPEC",&
            "Image and Spectrum have different OBSID/EXPID/INST/ORBIT")
    endif
    !
    call release (inSPEC_S) ! close the spectrum file
    !
    select case (instO(1:2))
    case ('M1','M2')
       prefix = "mos" // trim(adjustl(instO(2:2))) // trim(adjustl(expidO))
       nCCDs = 7
    case ('PN')
       prefix = "pn" // trim(adjustl(expidO))
       nCCDs = 4
    case default
       call error("INSTwrong",&
            "Instrument must be M1, M2, or PN only, check obj input file")
    end select
    !
    ! put spectrum and array dimensions check here??? (BP)
    ! IOW check 780x780, 2400, etc...
    ! also check returns from read_header_keywords elow vs elowO, etc!!!
    do i = 1, nCCDs
       ccds(i) = booleanParameter(ccdspar,i-1)
    end do
    ! delete this as unnecessary???!!! SAS-20???!!!
    if ( (instO .eq. 'M1') .and. revolutO .gt. 961) then
       if (ccds(6)) then
          call warning ("PostCCD6Loss",&
               "CCD#6 was selected but orbit after 961, ccds(6) set to F")
          ccds(6) = .FALSE.
       end if
    end if
    elow = intParameter(elowpar)
    write(celow,"(I5)") elow
    ehigh = intParameter(ehighpar)
    write(cehigh,"(I5)") ehigh
    spectrol = intParameter(spectrolpar)
    estring = "-" // trim(adjustl(celow)) // "-" // trim(adjustl(cehigh))
    if (spectrol .eq. 1) powerlaw = .TRUE.
    if (elow .ge. ehigh) then
       call error("elowGEehigh","Parameter elow must be less than ehigh")
    end if
    !
    if (powerlaw) then 
       pindex = realParameter(pindexpar)
       pnorm  = realParameter(pnormpar)
    else    
       bindl  = realParameter(bindlpar)
       bindh  = realParameter(bindhpar)
       break  = realParameter(bbreakpar)
       bnorm  = realParameter(bnormpar)
    endif
    ! user-specified, but if still 'default', generate filename
    outSPdetmap_F = stringParameter(outdetpar)
    if (trim(adjustl(outSPdetmap_F)) .eq. 'default') then
       outSPdetmap_F = trim(adjustl(prefix)) // "-protimdet" // &
            trim(adjustl(estring)) // ".fits"
    end if
    withsky = booleanParameter(withskypar)
    if (withsky) then
       outSPskymap_F = stringParameter(outskypar)
       if (trim(adjustl(outSPskymap_F)) .eq. 'default') then
          outSPskymap_F = trim(adjustl(prefix)) // "-protimsky" // &
               trim(adjustl(estring)) // ".fits"
       end if
    endif
    !
    ! SP map fits image is basically a clone of the FOV image
    ! insofar as header keywords
    !
    inquire(FILE=trim(outSPdetmap_F), EXIST=i_ok)
    if (i_ok) then
       if ( clobber() ) then
          write(errstr,*) "Output SP map file ",trim(adjustl(outSPdetmap_F)),&
               " exists and CLOBBER is set, will overwrite existing file."
          call message(AppMsg,Always)          
          call unlink (trim(outSPdetmap_F))
       else
          write(errstr,*) "Output SP map file ",trim(adjustl(outSPdetmap_F)),&
               " exists and CLOBBER is NOT set, terminating..."
          call error("noCLOBBER",trim(adjustl(errstr)))
       endif
    end if
    if (withsky) then 
       inquire(FILE=trim(outSPskymap_F), EXIST=i_ok)
       if (i_ok) then
          if ( clobber() ) then
             write(errstr,*) "SP map file ",trim(adjustl(outSPskymap_F)),&
                  " exists and CLOBBER is set, will overwrite existing file."
             call message(AppMsg,Always)          
             call unlink (trim(outSPskymap_F))
          else
             write(errstr,*) "SP map file ",trim(adjustl(outSPskymap_F)),&
                  " exists and CLOBBER is NOT set, terminating..."
             call error("noCLOBBER",trim(adjustl(errstr)))
          endif
       end if
    end if
    !
    call printParams (nCCDs,inIMG_F,inSPEC_F,outSPdetmap_F,outSPskymap_F, &
         celow,cehigh,powerlaw,ccds,withsky,pindex,pnorm,bindl,bindh,&
         break,bnorm)
    !
    ! input image array is bitpix = 32, 32-bit signed INTEGER
    !  output array image bitpix = -32, 32-bit FLOAT
    ! open output image file outSPskymap_F
    status = 0
    warndie = .TRUE. ! croak if not found or unreadable
    call check_open_set(outSPdetmap_F,outSPdetmap_S,3,status,warndie)
    ! 3 = CREATE, in gofio
    ! add the real32 SP map to output file
    outSPmap_A = addArray(outSPdetmap_S,"PRIMARY",real32,outIMGdim)
    outSPmap_RP => real32Array2Data(outSPmap_A)
    !
    ! clean these keywords up and add HISTORY/COMMENT lines!!!
    call copyAllAttributes(attributable(inIMG_A),attributable(outSPmap_A))
    call copyAllAttributes(attributable(inIMG_S),attributable(outSPdetmap_S))
    !
    ! cast input integer map to reals for output SP map
    outSPmap_RP = real(inIMG_JP)
    rn = sum(outSPmap_RP)
    ! the "before" map
    write(errstr,*) 'Sum of counts in input FOV image:  ',rn
    call message(AppMsg,Always)
    !

    !
    !  Set the bands to be included
    !
    ! eLo(6) = (/300.0, 750.0, 1250.0, 2000.0, 4000.0, 8000.0/)
    ! eHi(6) = (/749.0, 1249.0, 1999.0, 3999.0, 7999.0, 12000.0/)
    !
    doband = .false.
    scaleFactor = 1.
    j = 0
    do i = 1, 6
       if ( elow .gt. eHi(i) .and. ehigh .gt. eHi(i)) then
          ! both elow and ehigh above band range, just fill in til band selected reached
          doband(i) = .false. 
          scaleFactor(i) = 1.
          cycle
       endif
       if ( elow .ge. eLo(i) .and. ehigh .le. eHi(i) ) then ! both within single band
          doband(i) = .true.
          doband(i+1:6) = .false.
          scaleFactor(i) = (ehigh - elow) / (eHi(i) - eLo(i))
          scaleFactor(i+1:6) = 1.
          exit
       end if
       if ( elow .le. eLo(i) .and. ehigh .lt. eLo(i)) then
          ! both elow and ehigh below band range, fill in rest of bands and exit
          doband(i:6) = .false. 
          scaleFactor(i:6) = 1.
          exit
       endif
       ! all other possibilities, step thru
       do j=i,6
             if (ehigh .gt. eHi(j)) then
                doband(j) = .true.
                scaleFactor(j) = (eHi(j) - elow) / (eHi(j) - eLo(j))
                cycle
             else
                doband(j) = .true.
                scaleFactor(j) = (ehigh - eLo(j)) / (eHi(j) - eLo(j))
                exit                
             endif          
       end do
    end do
    !
    write(errstr,*) 'Bands and scale factors for the SP maps:'
    call message(AppMsg,Always)
    do i = 1, 6
       write(cc,"(I1)")i
       write(errstr,*) "Band(",trim(cc),") = ",doband(i),&
            " SF(",trim(cc),") = ",scaleFactor(i)
       call message(AppMsg,Always)
    end do
    nbands = count(doband)    
    !
    ! Loop through the selected SP bands
    ! bands are B1/B2/B3/B4/B5/B6
    ! energies are .7/.1/.2/.3/.4/.5/.6 
    ! filters are THIN/MEDIUM/THICK
    !
    do j=1, nbands ! number of bands selected
       do ff=1,3 ! filters THIN1/MEDIUM/THICK
          select case (ff)
          case (1)
             write(errstr,*) 'filter = Thin1','  band = ',bands(j)
             call message(AppMsg,Always)
             call CAL_setState(filterId=THIN1)
          case (2)
             write(errstr,*) 'filter = Medium',' band = ',bands(j)
             call message(AppMsg,Always)
             call CAL_setState(filterId=MEDIUM)
          case (3)
             write(errstr,*) 'filter = Thick','  band = ',bands(j)
             call message(AppMsg,Always)
             call CAL_setState(filterId=THICK)
          end select
          !
          write(errstr,*) "Current band to be retrieved by CAL_flaremap = ",bands(j)
          call message(AppMsg,Always)
          call CAL_flaremap(bands(j),flaremap_RP)
          !
          ! Mask chips if necessary (zero out chips not selected by user)
          ! the long code that was below has been moved to a subroutine
          ! in gofutils_mod.f90 in the task goflib
          !
          do i=1,nccds
             if (.not.ccds(i)) then
                call maskchips(instS,i,outSPmap_RP)
             end if
          end do
          !
          !  Co-add the data from all filters
          !
          print*,'summm ',sum(outSPmap_RP)
          do k=0,flareX-1
             do kk=0,flareY-1
               outSPmap_RP(k,kk) = outSPmap_RP(k,kk) + scaleFactor(j)*flaremap_RP(k,kk)
             enddo
          enddo
       enddo
    enddo
    !
    !  Find the number of counts in the detector map
    !
    rn = sum(outSPmap_RP)
    !
    !  Determine the number of SP counts in the band
    !
    eh = ehigh/1000.0
    el = elow/1000.0
    if(powerlaw) then
       rind = 1.0 - pindex
       SPrate = pnorm*(eh**rind - el**rind)/rind
    else
       splb = break**(-bindl)
       sphb = break**(-bindh)
       if(break .lt. 1.0) then
          bnorml = bnorm*sphb/splb
          bnormh = bnorm
       elseif (break .eq. 1.0) then
          bnorml = bnorm
          bnormh = bnorm
       else
          bnorml = bnorm
          bnormh = bnorm*splb/sphb
       endif
       if(eh .le. break) then
          rind = 1.0 - bindl
          SPrate = bnorm*(eh**rind - el**rind)/rind
       elseif(el .ge. break) then
          rind = 1.0 - bindh
          SPrate = bnorm*(break**(bindh-bindl))* &
               (eh**rind - el**rind)/rind
       else
          rind = 1.0 - bindl
          temp = bnorm*(break**rind - el**rind)/rind
          rind = 1.0 - bindh
          SPrate = temp + bnorm*(break**(bindh-bindl))* &
               (eh**rind - break**rind)/rind
       endif
    endif
    !
    !  Check to see whether MOS1 CCD#6 is missing
    ! remove???!!!
    if((inst .eq. 'M1') .and. (revolut .gt. 961)) then
       if( ccds(6) ) then
          call warning("PostCCD6Loss", &
               "Orbit is after MOS1 CCD6 lost, removing CCD6 from consideration.")
          ccds(6) = .FALSE.
          write(errstr,*) "Revised CCD selection:"
          call message(AppMsg,Always)
          do i = 1, 7
             write(cc,"(I1)")i
             write(errstr,*) "CCD(",trim(cc),") selected? (T/F)",ccds(i)
             call message(AppMsg,Always)
          end do
          i1 = 1
          i2 = 258
          ii1 = 257
          ii2 = 520
          do i=i1,i2
             do ii=ii1,ii2 ! check these???!!! should be i,ii?
                outSPmap_RP(i-1,ii-1) = 0.0
             enddo
          enddo
          rn = 0.0
          do i=1,flareX
             do ii=1,flareY ! check these???!!! should be i,ii?
                rn = rn + outSPmap_RP(i-1,ii-1)
             enddo
          enddo
       endif
       if(revolut .eq. 961) then
          call warning("CCD6LossOrbit",&
               "Observation took place during the observation where MOS1 CCD6 was lost.")
          write(errstr,*) &
               "Exposure U002 has CCD6 data, U003 does not"
          call message(AppMsg,Always)
       endif
    endif
    !
    !  Calculate the scale factor
    !
    SPrate = expoS * SPrate
    scale = SPrate / rn
    ! Multiply all elements by the scale factor
    outSPmap_RP = scale * outSPmap_RP
    !
    write(errstr,*) "Observation exposure  : ", expoS
    call message(AppMsg,Always)
    write(errstr,*) "Detector map counts   : ", rn
    call message(AppMsg,Always)
    write(errstr,*) "Proton flaring events : ", SPrate
    call message(AppMsg,Always)
    write(errstr,*) "Scale factor          : ", scale
    call message(AppMsg,Always)
    !
    ! get coordinate keywords from header to run rotimdet2sky
    !
    warndie = .false.
    call coord_keys (outSPdetmap_S,ra_pnt,dec_pnt,pa_pnt,crval,cdelt, &
         cunit,ctype,crpix,crota,radecsys,warndie)
    call release(outSPdetmap_S)
    logfile = "temppp.out"
    write(str_crval1,'(F12.5)')ra_pnt
    write(str_crval2,'(F12.5)')dec_pnt
    command = "esky2det datastyle=user " // "ra=" // trim(adjustl(str_crval1)) // &
         " dec=" // trim(adjustl(str_crval2)) // " outunit=det withheader=no" // &
         " calinfostyle=set calinfoset=" // trim(adjustl(outSPdetmap_F)) // &
         " verbosity=0 " // '>& ' // trim(adjustl(logfile))
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Always)
    call system(trim(adjustl(command)))
    !
    i_ok = .false.
    ios = 0
    inquire(FILE=trim(logfile), EXIST=i_ok)
    if (i_ok) then
       write(errstr,*) "Ran esky2det, opening and reading dxcen/dycen from log: ", &
            trim(adjustl(logfile))
       call message(AppMsg,Noisy)
       lun = GetLun() ! in goflib/gofutils
       ios = 0
       open (lun,file=trim(logfile), iostat=ios)
       if (ios .ne. 0) then
          call error ("noLUN","Cannot get LUN for esky2det text output file")
       end if
       do ! esky2det dumps crap into temppp.out text file ???!!!
          read(lun,'(A)',iostat=reason) line
          if (reason > 0)  then
             call error ("badTEXTread","Error reading output temppp.out esky2det file")
          else if (reason < 0) then ! EOF
             exit
          else
             crap = 0
             crap = scan(line,'abcdefghijklmnopqrstuvwxyz')
             if (crap .gt. 0) then
                cycle
             else
                read(line,*)dxcen,dycen
                exit
             end if
          end if
       end do
       write(errstr,*) "DXCEN/DYCEN from esky2det: ",dxcen,dycen
       call message(AppMsg,Always)
       call unlink ("temppp.out") ! remove temp file
    else
       write(errstr,*)"esky2det temp output file (temppp.out) cannot be read "
       call error("badEsky2DetRun",trim(errstr))
    end if
    !
    if (withsky) then
       write(str_detx,'(F12.5)')dxcen
       write(str_dety,'(F12.5)')dycen
       command = "rotdet2sky intemplate=" // trim(adjustl(inIMG_F)) // &
            " inimage=" // trim(adjustl(outSPdetmap_F)) // &
            " outimage=" // trim(adjustl(outSPskymap_F)) // &
            " withmask=F " // & 
            " withskyxy=T" // &
            " detx=" // trim(adjustl(str_detx)) // &
            " dety=" // trim(adjustl(str_dety)) // &
            " skyx=" // trim(adjustl(str_crval1)) // & 
            " skyy=" // trim(adjustl(str_crval2))
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Always)
!       print*,trim(adjustl(command))
       call system(trim(adjustl(command)))
       call addlohi(outSPskymap_F,elow,ehigh)
    end if

    call addlohi(outSPdetmap_F,elow,ehigh)

    ! add history/comment keywords???!!!
    
    return

  end subroutine proton

  subroutine printParams (nCCDs,inIMG_F,inSPEC_F,outSPdetmap_F,outSPskymap_F, &
       celow, cehigh, powerlaw, ccds, withsky, pindex, pnorm, bindl, bindh, &
       break, bnorm)

    character(len=*), intent(in) :: inIMG_F, inSPEC_F, outSPdetmap_F, &
         outSPskymap_F, celow, cehigh
    character(len=5) :: foo
    character(len=1) :: cc
    integer :: i, nCCDs
    logical, intent(in) :: ccds(7), withsky, powerlaw
    real, intent(in) :: pindex, pnorm, bindl, bindh, break, bnorm
    
    write(errstr,*) "Input parameters:"
    call message(AppMsg,Always)
    write(errstr,*) "Input Spectral Filename     : ",trim((adjustl(inSPEC_F)))
    call message(AppMsg,Always)
    write(errstr,*) "Input Image Filename        : ",trim(adjustl(inIMG_F))
    call message(AppMsg,Always)
    write(errstr,*) "Output SP Map (det) Filename: ",trim(adjustl(outSPdetmap_F))
    call message(AppMsg,Always)
    if (withsky) then 
       write(errstr,*) "Output SP Map (sky) Filename: ",trim(adjustl(outSPskymap_F))
       call message(AppMsg,Always)
    endif
    write(errstr,*) "Selected energy lower limit : ",trim(adjustl(celow))
    call message(AppMsg,Always)
    write(errstr,*) "Selected energy upper limit : ",trim(adjustl(cehigh))
    call message(AppMsg,Always)
    if (powerlaw) then
       write(errstr,*) "Spectral control is power law: "
       call message(AppMsg,Always)
       write(errstr,*) "Photon index             : ",pindex
       call message(AppMsg,Always)
       write(errstr,*) "Photon normalization     : ",pnorm
       call message(AppMsg,Always)
    else 
       write(errstr,*) "Spectral control is broken power law: "
       call message(AppMsg,Always)
       write(errstr,*) "Broken Power law low-energy index   : ",bindl
       call message(AppMsg,Always)
       write(errstr,*) "Broken Power law break energy (keV) : ",break
       call message(AppMsg,Always)
       write(errstr,*) "Broken Power law high-energy index  : ",bindh
       call message(AppMsg,Always)
       write(errstr,*) "Broken Power law normalization      : ",bnorm
       call message(AppMsg,Always)
    end if
    foo = "CCD("
    if (nCCDS .eq. 4) foo = "QUAD("
    do i = 1, nCCDs
       write(cc,"(I1)")i
       write(errstr,*) trim(adjustl(foo)),cc,") selected? (T/F)",ccds(i)
       call message(AppMsg,Always)
    end do

    return
  end subroutine printParams

end module proton_mod
